import Foundation

struct YearlyCount: Codable, Identifiable, Hashable, Sendable {
    var id: UUID = UUID()
    var year: Int = 0
    var count: Int = 0
    var typeIdentifier: String = ""
    var isApproximate: Bool = false
    init(year: Int, count: Int, typeIdentifier: String, isApproximate: Bool = false) {
        self.id = UUID()
        self.year = year
        self.count = count
        self.typeIdentifier = typeIdentifier
        self.isApproximate = isApproximate
    }
}
